import 'rxjs/add/operator/toPromise';
import { AsyncSubject as AsyncSubject$1 } from 'rxjs/AsyncSubject';
import { Observable as Observable$1 } from 'rxjs/Observable';
import { Inject, InjectionToken, Injector, NgModule, Optional, Pipe, isDevMode } from '@angular/core';
import { BehaviorSubject as BehaviorSubject$1 } from 'rxjs/BehaviorSubject';
import { ReplaySubject as ReplaySubject$1 } from 'rxjs/ReplaySubject';
import { Subject as Subject$1 } from 'rxjs/Subject';
import { CommonModule } from '@angular/common';
import { TestBed } from '@angular/core/testing';
import { createStore } from 'redux';

class ActionInterface {
}

class MetadataManager {
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_KEY, MetadataManager.ACTION_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setActionContextMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.ACTION_CONTEXT_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getActionContextMetadata(target) {
        return MetadataManager.get(target, MetadataManager.ACTION_CONTEXT_KEY, MetadataManager.ACTION_CONTEXT_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setReducerMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.REDUCER_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getReducerMetadata(target) {
        return MetadataManager.get(target, MetadataManager.REDUCER_KEY, MetadataManager.REDUCER_DEFAULT);
    }
    /**
     * @template T
     * @param {?} target
     * @param {?} data
     * @return {?}
     */
    static setStateMetadata(target, data) {
        MetadataManager.set(target, MetadataManager.STATE_KEY, data);
    }
    /**
     * @template T
     * @param {?} target
     * @return {?}
     */
    static getStateMetadata(target) {
        return MetadataManager.get(target, MetadataManager.STATE_KEY, MetadataManager.STATE_DEFAULT);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} data
     * @return {?}
     */
    static set(target, key, data) {
        Reflect['defineMetadata'](key, data, target);
    }
    /**
     * @template D, T
     * @param {?} target
     * @param {?} key
     * @param {?} defaultData
     * @return {?}
     */
    static get(target, key, defaultData) {
        defaultData = defaultData || /** @type {?} */ ({});
        return Object.assign({}, defaultData, Reflect['getMetadata'](key, target));
    }
}
MetadataManager.ACTION_KEY = Symbol('@ReduxAction');
MetadataManager.ACTION_DEFAULT = {
    contextClazz: null,
    type: '',
};
MetadataManager.ACTION_CONTEXT_KEY = Symbol('@ReduxActionContext');
MetadataManager.ACTION_CONTEXT_DEFAULT = {
    prefix: '',
};
MetadataManager.REDUCER_KEY = Symbol('@ReduxReducer');
MetadataManager.REDUCER_DEFAULT = {
    reducers: [],
};
MetadataManager.STATE_KEY = Symbol('@ReduxState');
MetadataManager.STATE_DEFAULT = {
    name: '',
};

/**
 * @param {?} config
 * @return {?}
 */
function ReduxActionContext(config) {
    return (constructor) => {
        MetadataManager.setActionContextMetadata(constructor, config);
        return constructor;
    };
}

class Registry {
    /**
     * @return {?}
     */
    static reset() {
        Registry._store = new AsyncSubject$1();
        Registry._reducers = [];
    }
    /**
     * @param {?} store
     * @return {?}
     */
    static registerStore(store) {
        Registry._store.next(store);
        Registry._store.complete();
    }
    /**
     * @param {?} stateName
     * @param {?} types
     * @param {?} reducer
     * @return {?}
     */
    static registerReducer(stateName, types, reducer) {
        if (Array.isArray(types)) {
            types.forEach((type) => {
                Registry._reducers.push({
                    reducer,
                    stateName,
                    type,
                });
            });
        }
        else {
            Registry._reducers.push({
                reducer,
                stateName,
                type: /** @type {?} */ (types),
            });
        }
    }
    /**
     * @param {?} state
     * @return {?}
     */
    static registerState(state) {
        Registry.getStore().then((store) => {
            const /** @type {?} */ stateConfig = MetadataManager.getStateMetadata(state.constructor);
            const /** @type {?} */ initState = state.getInitialState();
            let /** @type {?} */ initStateToResolve = initState;
            if (initState instanceof Observable$1) {
                initStateToResolve = initState.toPromise();
            }
            Promise.resolve(initStateToResolve).then((initialState) => {
                store.dispatch({
                    payload: {
                        initialValue: initialState,
                        name: stateConfig.name,
                    },
                    type: Registry.ACTION_REGISTER_STATE,
                });
            });
        });
    }
    /**
     * @param {?} type
     * @return {?}
     */
    static getReducerItemsByType(type) {
        return Registry._reducers
            .filter(reducerItem => getActionTypeByFunction(reducerItem.type) === type);
    }
    /**
     * @return {?}
     */
    static getStore() {
        return new Promise(Registry._store.subscribe.bind(Registry._store));
    }
}
Registry.ACTION_REGISTER_STATE = `ngx-redux://registerState`;
Registry._store = new AsyncSubject$1();
Registry._reducers = [];

/**
 * @param {?=} config
 * @return {?}
 */
function ReduxAction(config) {
    return (target, propertyKey, descriptor) => {
        config = Object.assign({
            type: propertyKey,
        }, config || {});
        const /** @type {?} */ originalFunction = descriptor.value;
        const /** @type {?} */ proxyFunction = function () {
            const /** @type {?} */ contextData = MetadataManager.getActionContextMetadata(target.constructor);
            const /** @type {?} */ type = getActionType(contextData.prefix, config.type);
            const /** @type {?} */ returnValue = originalFunction.apply(this, arguments);
            Promise.all([
                Promise.resolve(returnValue),
                Registry.getStore(),
            ]).then(([payload, store]) => {
                store.dispatch({ payload, type });
            });
            return returnValue;
        };
        MetadataManager.setActionMetadata(proxyFunction, {
            contextClazz: target.constructor,
            type: config.type,
        });
        descriptor.value = proxyFunction;
    };
}
/**
 * @param {?} target
 * @return {?}
 */
function getActionTypeByFunction(target) {
    const { type, contextClazz } = MetadataManager.getActionMetadata(target);
    const { prefix } = MetadataManager.getActionContextMetadata(contextClazz);
    return getActionType(prefix, type);
}
/**
 * @param {...?} rest
 * @return {?}
 */
function getActionType(...rest) {
    return rest.filter((i) => i.length > 0).join('');
}

class ReduxModuleRootReducer {
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduce(state, action) {
        if (action.type === Registry.ACTION_REGISTER_STATE) {
            return ReduxModuleRootReducer.reduceRegisterState(state, /** @type {?} */ (action));
        }
        return ReduxModuleRootReducer.reduceByRegisteredReducers(state, action);
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceRegisterState(state, action) {
        const { name, initialValue } = action.payload;
        return Object.assign({}, state, {
            [name]: initialValue,
        });
    }
    /**
     * @param {?} state
     * @param {?} action
     * @return {?}
     */
    static reduceByRegisteredReducers(state, action) {
        const /** @type {?} */ reducers = Registry.getReducerItemsByType(action.type);
        return reducers.reduce((stateToReduce, reducerItem) => {
            const { stateName, reducer } = reducerItem;
            return Object.assign({}, stateToReduce, {
                [stateName]: reducer(stateToReduce[stateName], action),
            });
        }, state);
    }
}

/**
 * @template S, P
 * @param {?} types
 * @return {?}
 */
function ReduxReducer(types) {
    return (target, propertyKey) => {
        const /** @type {?} */ data = MetadataManager.getReducerMetadata(target.constructor);
        data.reducers.push({ types, reducer: target[propertyKey].bind(target) });
        MetadataManager.setReducerMetadata(target.constructor, data);
        return target[propertyKey];
    };
}

class StateDefToken extends InjectionToken {
    constructor() {
        super('StateDefToken');
    }
}

let ReduxStateSelectorSubjectType = {};
ReduxStateSelectorSubjectType.DEFAULT = 0;
ReduxStateSelectorSubjectType.SUBJECT = 1;
ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT = 2;
ReduxStateSelectorSubjectType.REPLAY_SUBJECT = 3;
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.DEFAULT] = "DEFAULT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.SUBJECT] = "SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT] = "BEHAVIOR_SUBJECT";
ReduxStateSelectorSubjectType[ReduxStateSelectorSubjectType.REPLAY_SUBJECT] = "REPLAY_SUBJECT";

class ReduxStateSelector {
    /**
     * @param {?} expression
     * @param {?=} context
     */
    constructor(expression, context) {
        this.expression = expression;
        this.context = context;
        if (!expression.startsWith(ReduxStateSelector.DELIMITER)) {
            if (!context) {
                throw new Error('You need to provide a state context if you want to use relativ selectors');
            }
            this.expression = `/${MetadataManager.getStateMetadata(context).name}/${this.expression}`;
        }
    }
    /**
     * @template S
     * @return {?}
     */
    getObservable() {
        return this.getSubject().asObservable();
    }
    /**
     * @template S
     * @param {?} initialValue
     * @return {?}
     */
    getBehaviorSubject(initialValue) {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.BEHAVIOR_SUBJECT, initialValue));
    }
    /**
     * @template S
     * @return {?}
     */
    getReplaySubject() {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.REPLAY_SUBJECT));
    }
    /**
     * @template S
     * @return {?}
     */
    getSubject() {
        return /** @type {?} */ (this.getBySubjectType(ReduxStateSelectorSubjectType.SUBJECT));
    }
    /**
     * @template S
     * @param {?} type
     * @param {?=} initialValue
     * @return {?}
     */
    getBySubjectType(type, initialValue) {
        let /** @type {?} */ subject;
        switch (type) {
            case ReduxStateSelectorSubjectType.REPLAY_SUBJECT:
                subject = new ReplaySubject$1();
                break;
            case ReduxStateSelectorSubjectType.SUBJECT:
                subject = new Subject$1();
                break;
            default:
                subject = new BehaviorSubject$1(initialValue || null);
        }
        Registry.getStore().then(store => {
            subject.next(this.getValueFromState(store.getState()));
            store.subscribe(() => subject.next(this.getValueFromState(store.getState())));
        });
        return subject;
    }
    /**
     * @template T, S
     * @param {?} state
     * @return {?}
     */
    getValueFromState(state) {
        return this.expression.split(ReduxStateSelector.DELIMITER)
            .filter(ReduxStateSelector.isPropertyKeyValid)
            .reduce((previousValue, propertyKey) => {
            if (!previousValue || !previousValue[propertyKey]) {
                return null;
            }
            return previousValue[propertyKey];
        }, state);
    }
    /**
     * @param {?} propertyKey
     * @return {?}
     */
    static isPropertyKeyValid(propertyKey) {
        return propertyKey !== '';
    }
}
ReduxStateSelector.DELIMITER = '/';

class ReduxSelectPipe {
    /**
     * @param {?} stateDef
     */
    constructor(stateDef) {
        this.stateDef = stateDef;
    }
    /**
     * @param {?} selector
     * @return {?}
     */
    transform(selector) {
        return new ReduxStateSelector(selector, this.stateDef.provider).getObservable();
    }
}
ReduxSelectPipe.decorators = [
    { type: Pipe, args: [{ name: 'reduxSelect' },] },
];
/**
 * @nocollapse
 */
ReduxSelectPipe.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [StateDefToken,] },] },
];

/**
 * @param {?} expression
 * @param {?=} context
 * @return {?}
 */
function ReduxSelect(expression, context) {
    return (target, propertyKey) => {
        let /** @type {?} */ selector;
        Object.defineProperty(target, propertyKey, {
            configurable: true,
            enumerable: true,
            get: () => {
                if (!selector) {
                    selector = new ReduxStateSelector(expression, context);
                }
                return selector.getObservable();
            },
        });
    };
}

/**
 * @param {?} config
 * @return {?}
 */
function ReduxState(config) {
    return (constructor) => {
        MetadataManager.setStateMetadata(constructor, config);
        return constructor;
    };
}

class ReduxStore extends InjectionToken {
    constructor() {
        super('ReduxStore');
    }
}

class StateDefinitionManager {
    /**
     * @param {?} stateDef
     * @return {?}
     */
    static registerReducers(stateDef) {
        if (stateDef.reducers) {
            const /** @type {?} */ stateMetadata = MetadataManager.getStateMetadata(stateDef.provider);
            stateDef.reducers
                .map((reducer) => MetadataManager.getReducerMetadata(reducer.constructor))
                .forEach((metadata) => {
                metadata.reducers.forEach(reducer => {
                    Registry.registerReducer(stateMetadata.name, reducer.types, reducer.reducer);
                });
            });
        }
    }
}

class ReduxTestingModule {
    constructor() {
        Registry.reset();
        Registry.registerStore(ReduxTestingModule.storeFactory());
    }
    /**
     * @param {?} stateDef
     * @return {?}
     */
    static addStateDefinition(stateDef) {
        Registry.registerState(TestBed.get(stateDef.provider));
        StateDefinitionManager.registerReducers(stateDef);
    }
    /**
     * @return {?}
     */
    static storeFactory() {
        return createStore(ReduxModuleRootReducer.reduce, {});
    }
}
ReduxTestingModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxTestingModule.ctorParameters = () => [];

class ReduxModule {
    /**
     * @param {?=} stateDefs
     * @param {?=} store
     * @param {?=} injector
     */
    constructor(stateDefs = [], store = null, injector) {
        this.injector = injector;
        if (store) {
            Registry.registerStore(store);
        }
        stateDefs
            .filter(stateDef => !!stateDef)
            .map(stateDef => {
            return {
                metadata: MetadataManager.getStateMetadata(stateDef.provider),
                stateDef,
            };
        })
            .filter(({ metadata }) => !ReduxModule.knownStateDefinitions[metadata.name])
            .forEach(({ metadata, stateDef }) => {
            this.initState(stateDef);
            ReduxModule.knownStateDefinitions[metadata.name] = stateDef;
        });
    }
    /**
     * @param {?} stateDef
     * @return {?}
     */
    initState(stateDef) {
        Registry.registerState(this.injector.get(stateDef.provider));
        StateDefinitionManager.registerReducers(stateDef);
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forChild(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: [
                { provide: StateDefToken, useValue: config.state || null, multi: true },
                config.state ? config.state.provider : null,
            ],
        };
    }
    /**
     * @param {?=} config
     * @return {?}
     */
    static forRoot(config = {}) {
        return {
            ngModule: ReduxModule,
            providers: config.state ? [
                { provide: ReduxStore, useFactory: config.storeFactory || ReduxModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
                config.state ? config.state.provider : null,
            ] : [
                { provide: ReduxStore, useFactory: config.storeFactory || ReduxModule.defaultStoreFactory },
                { provide: StateDefToken, useValue: config.state || null, multi: true },
            ],
        };
    }
    /**
     * @return {?}
     */
    static defaultStoreFactory() {
        return createStore(ReduxModuleRootReducer.reduce, {}, ReduxModule.defaultEnhancerFactory());
    }
    /**
     * @return {?}
     */
    static defaultEnhancerFactory() {
        if (console && window['__REDUX_DEVTOOLS_EXTENSION__'] && isDevMode()) {
            return window['__REDUX_DEVTOOLS_EXTENSION__']();
        }
        return ReduxModule.nullEnhancer;
    }
    /**
     * @param {?} next
     * @return {?}
     */
    static nullEnhancer(next) {
        return next;
    }
}
ReduxModule.knownStateDefinitions = {};
ReduxModule.decorators = [
    { type: NgModule, args: [{
                declarations: [
                    ReduxSelectPipe,
                ],
                exports: [
                    ReduxSelectPipe,
                ],
                imports: [
                    CommonModule,
                ],
            },] },
];
/**
 * @nocollapse
 */
ReduxModule.ctorParameters = () => [
    { type: Array, decorators: [{ type: Optional }, { type: Inject, args: [StateDefToken,] },] },
    { type: undefined, decorators: [{ type: Optional }, { type: Inject, args: [ReduxStore,] },] },
    { type: Injector, },
];

/**
 * Generated bundle index. Do not edit.
 */

export { ActionInterface, ReduxActionContext, ReduxAction, getActionTypeByFunction, getActionType, ReduxModuleRootReducer, ReduxReducer, ReduxSelect, ReduxSelectPipe, ReduxStateSelector, ReduxStateSelectorSubjectType, ReduxState, ReduxStore, ReduxTestingModule, ReduxModule, StateDefToken as ɵa };
//# sourceMappingURL=ngx-redux-core.js.map
